/*
** Module   : NetSocket.java
** Abstract : combines a socket and related input/output streams into a
**            manageable unit
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 NVS 20050308   @20245 Created. This class extends and replaces
**                           SSLSocket in net and security packages.
** 002 NVS 20050311   @20287 Simplified by removing DataInput and Output
**                           streams. Authentication now uses ObjectInput
**                           and Output streams as well.
** 003 GES 20081103   @40327 Rework to remove unnecessary features and
**                           to allow non-secure sockets to be used.
** 004 GES 20090609   @42628 Provided access to local socket info.
** 005 OM  20131022          Added a list of SessionListeners and minimum management methods.
*/ 

package com.goldencode.p2j.net;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.net.ssl.*;

/**
 * This class represents a secure socket and all related input/output streams
 * as a one unit. This solves the problem of temporary streams closing other
 * streams and the underlying socket when finalized.
 */
public class NetSocket
{
   /** Embedded socket which may or may not be secure. */
   private Socket socket = null;

   /** SSL session. */
   private SSLSession session = null;

   /** Object input stream. */
   private ObjectInputStream objectInp = null;
   
   /** Object output stream. */
   private ObjectOutputStream objectOut = null;

   /** Stores a list of <code>SessionListener</code>s */
   private ArrayList<SessionListener> sessionListeners = null;
   
   /**
    * Constructor
    *
    * @param    socket
    *           A connected socket to embed into this class. This may be a
    *           secure socket but it is not required.
    *
    * @throws   IOException
    *           If one of the streams being built on top of the socket throws
    *           the same exception.
    */
   public NetSocket(Socket socket)
   throws IOException
   {
      this.socket = socket;

      if (socket instanceof SSLSocket)
      {
         SSLSocket ssl = (SSLSocket) socket;
         
         // get the SSL session, if this is a secure socket (NOTE: this will
         // cause the SSL handshake to occur if it has not already!)
         session = ssl.getSession();
      }

      // create object streams
      objectOut = new ObjectOutputStream(socket.getOutputStream());
      objectInp = new ObjectInputStream(socket.getInputStream());
   }
   
   /**
    * Gets the SSL session.
    *
    * @return   The SSL session or <code>null</code> if the socket is not
    *           secure.
    */
   public SSLSession getSession()
   {
      return session;
   }

   /**
    * Gets the object input stream.
    *
    * @return   The embedded object input stream.
    */
   public ObjectInputStream getObjectInput()
   {
      return objectInp;
   }

   /**
    * Gets the object output stream.
    *
    * @return   The embedded object output stream.
    */
   public ObjectOutputStream getObjectOutput()
   {
      return objectOut;
   }
   
   /**
    * Closes the socket. 
    */
   public void close()
   {
      try
      {
         socket.close();
      }
      catch (IOException ex)
      {
         // best efforts
      }
   }

   /**
    * Reports if the socket is closed (and thus can no longer be used).
    *
    * @return   <code>true</code> if the socket is closed. 
    */
   public boolean isClosed()
   {
      return socket.isClosed();
   }
   
   /**
    * Access the address of the local system, through which we are connecting.
    *
    * @return   The local address.
    */
   public InetAddress getLocalAddr()
   {
      return socket.getLocalAddress();
   }
   
   /**
    * Access the address of the local system, through which we are connecting.
    *
    * @return   The local socket address.
    */
   public InetSocketAddress getLocalSockAddr()
   {
      return new InetSocketAddress(socket.getLocalAddress(), socket.getLocalPort());
   }
   
   /**
    * Access the address of the remote system, to which we are connecting.
    *
    * @return   The remote address.
    */
   public InetAddress getRemoteAddr()
   {
      return socket.getInetAddress();
   }
   
   /**
    * Access the address of the remote system, to which we are connecting.
    *
    * @return   The remote socket address.
    */
   public InetSocketAddress getRemoteSockAddr()
   {
      return new InetSocketAddress(socket.getInetAddress(), socket.getPort());
   }

   /**
    * Gets the list with added session listeners. May return <code>null</code> if none has
    * been added.
    *
    * @return  The current session listener.
    */
   public List<SessionListener> getSessionListeners()
   {
      return sessionListeners;
   }

   /**
    * Adds a session listener.
    *
    * @param   sessionListener
    *          A session listener to be added.
    */
   public void addSessionListener(SessionListener sessionListener)
   {
      if (sessionListeners == null)
      {
         sessionListeners = new ArrayList<>();
      }

      sessionListeners.add(sessionListener);
   }
}
